<?php

namespace Drupal\tmgmt_blackbird\Rest;

use Drupal\Core\Access\AccessResult;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\language\ConfigurableLanguageManagerInterface;
use Drupal\tmgmt\Entity\Translator;
use Drupal\tmgmt\JobInterface;
use Drupal\tmgmt_file\Format\FormatManager;
use Symfony\Component\HttpFoundation\JsonResponse;
use Symfony\Component\HttpFoundation\RequestStack;
use Symfony\Component\HttpFoundation\Response;
use Symfony\Component\HttpKernel\Exception\AccessDeniedHttpException;
use Symfony\Component\HttpKernel\Exception\BadRequestHttpException;

/**
 * Service to handle Blackbird API requests.
 */
class BlackbirdApi {

  /**
   * The request stack.
   *
   * @var \Symfony\Component\HttpFoundation\RequestStack
   */
  protected RequestStack $requestStack;

  /**
   * The entity type manager.
   *
   * @var \Drupal\Core\Entity\EntityTypeManagerInterface
   */
  protected EntityTypeManagerInterface $entityTypeManager;

  /**
   * The language manager.
   *
   * @var \Drupal\language\ConfigurableLanguageManagerInterface
   */
  protected ConfigurableLanguageManagerInterface $languageManager;

  /**
   * The format manager.
   *
   * @var \Drupal\tmgmt_file\Format\FormatManager
   */
  protected FormatManager $formatManager;

  /**
   * Constructs a BlackbirdApi instance.
   *
   * @param \Symfony\Component\HttpFoundation\RequestStack $request_stack
   *   The request stack.
   * @param \Drupal\Core\Entity\EntityTypeManagerInterface $entity_type_manager
   *   The entity type manager.
   * @param \Drupal\language\ConfigurableLanguageManagerInterface $language_manager
   *   The language manager.
   * @param \Drupal\tmgmt_file\Format\FormatManager $format_manager
   *   The format manager.
   */
  public function __construct(RequestStack $request_stack, EntityTypeManagerInterface $entity_type_manager, ConfigurableLanguageManagerInterface $language_manager, FormatManager $format_manager) {
    $this->requestStack = $request_stack;
    $this->entityTypeManager = $entity_type_manager;
    $this->languageManager = $language_manager;
    $this->formatManager = $format_manager;
  }

  /**
   * Returns a list of available languages.
   *
   * @return \Symfony\Component\HttpFoundation\JsonResponse
   *   The JSON response containing the available languages.
   */
  public function languages(): JsonResponse {
    return new JsonResponse(array_map(static function ($language) {
      return [
        'id' => $language->getId(),
        'name' => $language->getName(),
      ];
    }, $this->languageManager->getLanguages()));
  }

  /**
   * Returns a list of translation jobs for Blackbird.
   *
   * @return \Symfony\Component\HttpFoundation\JsonResponse
   *   The JSON response containing the translation jobs.
   *
   * @throws \Symfony\Component\HttpKernel\Exception\BadRequestHttpException
   *   Thrown when the request contains invalid parameters.
   */
  public function jobs(): JsonResponse {
    $translator = $this->getTranslatorForRequest();

    // Fetch the query parameters for the request to allow filtering the jobs.
    $params = $this->requestStack->getCurrentRequest()->query->all();

    // Build the query to fetch jobs for the current Blackbird translator.
    $query = $this->entityTypeManager->getStorage('tmgmt_job')->getQuery()
      ->condition('translator', $translator->id())
      ->accessCheck(FALSE)
      ->sort('created');

    // Allow jobs to be filtered by created date. We only return jobs created
    // after the provided date.
    if (!empty($params['created'])) {
      if (!is_numeric($params['created'])) {
        throw new BadRequestHttpException('Invalid created parameter. The created parameter must be a valid timestamp.');
      }
      $query->condition('created', $params['created'], '>=');
    }

    // Allow the jobs to be filtered by source language.
    if (!empty($params['source'])) {
      $query->condition('source_language', $params['source']);
    }

    // Allow the jobs to be filtered by target language.
    if (!empty($params['target'])) {
      $query->condition('target_language', $params['target']);
    }

    // Allow the jobs to be filtered by state. We default to active jobs. Make
    // sure to only allow valid 'state' parameters.
    $valid_states = [
      'active' => JobInterface::STATE_ACTIVE,
      'aborted' => JobInterface::STATE_ABORTED,
      'completed' => JobInterface::STATE_FINISHED,
    ];
    if (!empty($params['state']) && !array_key_exists($params['state'], $valid_states)) {
      throw new BadRequestHttpException('Invalid state parameter. The state parameter must be one of: ' . implode(', ', array_keys($valid_states)) . '.');
    }
    $query->condition('state', $valid_states[$params['state']] ?? JobInterface::STATE_ACTIVE);

    // Fetch all active jobs for the current Blackbird translator.
    $job_ids = $query->execute();
    /** @var \Drupal\tmgmt\JobInterface[] $jobs */
    $jobs = $this->entityTypeManager->getStorage('tmgmt_job')->loadMultiple($job_ids);

    // Fetch the job entities.
    return new JsonResponse(array_map(static function ($job) {
      return [
        'id' => $job->id(),
        'name' => $job->label(),
        'source' => $job->getSourceLangcode(),
        'target' => $job->getTargetLangcode(),
        'created' => $job->getCreatedTime(),
      ];
    }, $jobs));
  }

  /**
   * Handles the export or import of the HTML for a translation job.
   *
   * @param \Drupal\tmgmt\JobInterface $tmgmt_job
   *   The translation job.
   *
   * @return \Symfony\Component\HttpFoundation\Response
   *   A response containing the translation job HTML output.
   */
  public function job(JobInterface $tmgmt_job): Response {
    $translator = $this->getTranslatorForRequest();

    // Check if the translator is allowed to access the translation job.
    if ($tmgmt_job->getTranslator()->id() !== $translator->id()) {
      throw new AccessDeniedHttpException('The translator is not allowed to access the translation job.');
    }

    // Fetch the current request to determine the request method.
    $request = $this->requestStack->getCurrentRequest();
    if (!$request) {
      throw new BadRequestHttpException('Invalid request.');
    }

    /** @var \Drupal\tmgmt_file\Format\FormatInterface $html_plugin */
    $html_plugin = $this->formatManager->createInstance('html');

    // Check if the request is a GET request to export the HTML.
    if ($request->isMethod('GET')) {
      $response = new Response();
      $response->setContent($html_plugin->export($tmgmt_job));
      $response->headers->set('Content-Type', 'text/html');
      return $response;
    }

    // Check if the request is a POST request to import the HTML.
    if ($request->isMethod('POST')) {
      $html_content = $request->getContent();

      // Validate the HTML content before importing. The validateImport method
      // currently needs a file to validate the input. We write a temporary file
      // to validate the input for now. This can be removed once #3484954 is
      // resolved.
      // @see https://www.drupal.org/project/tmgmt/issues/3484954
      $path = 'private://blackbird_import_' . $tmgmt_job->id() . '.html';
      file_put_contents($path, $html_content);
      $job = $html_plugin->validateImport($path);
      unlink($path);

      // We ignore the phpstan error here because the validateImport method
      // returns FALSE if the input could not be validated.
      // @phpstan-ignore-next-line
      if (!$job instanceof JobInterface) {
        throw new BadRequestHttpException('Invalid HTML content.');
      }

      // When the translation job has already been completed, we can not import
      // new translations.
      if ($job->isFinished()) {
        throw new BadRequestHttpException('The translation job has already been completed.');
      }

      // We are now ready to import the HTML content.
      try {
        $job->addTranslatedData($html_plugin->import($html_content, FALSE));
      }
      catch (\Exception $e) {
        throw new BadRequestHttpException('An error occured while importing the translations.');
      }

      // Return a 200 OK response to indicate that the import was successful.
      return new Response();
    }

    // Return a 405 Method Not Allowed response for unsupported request methods.
    return new Response(NULL, Response::HTTP_METHOD_NOT_ALLOWED);
  }

  /**
   * Checks access to the API.
   *
   * @return \Drupal\Core\Access\AccessResult
   *   The access result.
   */
  public function access(): AccessResult {
    // Check if the request contains a valid API key.
    $api_key = $this->requestStack->getCurrentRequest()->headers->get('x-api-key');
    if (!$api_key) {
      return AccessResult::forbidden('No API key provided.');
    }

    // Fetch the Blackbird translator entity for the qurrent request.
    $translator = $this->getTranslatorForRequest();

    // Check if the API key is valid.
    if (!$translator instanceof Translator) {
      return AccessResult::forbidden('Invalid API key.');
    }

    return AccessResult::allowed()->addCacheableDependency($translator);
  }

  /**
   * Returns the translator for the current request.
   *
   * @return \Drupal\tmgmt\Entity\Translator|null
   *   The translator entity or NULL if not found.
   */
  protected function getTranslatorForRequest(): ?Translator {
    // Fetch the Blackbird translator config entity for the provides API key.
    /** @var \Drupal\tmgmt\Entity\Translator[] $translators */
    $translators = $this->entityTypeManager->getStorage('tmgmt_translator')->loadByProperties([
      'plugin' => 'blackbird',
      'settings.api_key' => $this->requestStack->getCurrentRequest()->headers->get('x-api-key'),
    ]);

    // Return the first translator plugin that matches the API key.
    return reset($translators);
  }

}
